"""
MXS is the Math eXploration Solver.
Copyright (C) 2022  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""


def get_pv_params():
    pattern = input("What is the pattern?  ")
    char_per_line = input("How many characters on a line?  ")
    target_line = input("What line number are you interested in?  ")
    target_char = input(f"What character on line {target_line} are you interested in?  ")

    return pattern, char_per_line, target_line, target_char


def reveal_position_value(pattern, char_per_line, target_line, target_char):
    char_per_line = int(char_per_line)
    target_line = int(target_line)
    target_char = int(target_char)

    target_char_delta = char_per_line - target_char
    target_position = (target_line * char_per_line) - target_char_delta

    remainder = target_position % len(pattern)
    value = pattern[remainder - 1]

    return target_position, value



"""
MXS is the Math eXploration Solver.
Copyright (C) 2022  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com

The error module provides custom LED Error Classes, as well as an ExitCode
class that supplies exit code values used when exiting the LED program.
Errror messages are available in the LED log file, which is located at
~/.local/log/led.log by default.

In bash, the exit code can be retrieved in the usual manner, by querying
$? after the program exits.

Exit codes for many (most) systems are usually 0 for success and 1 for
general failure.

BSD has attempted to standardize exit codes. These values are usually defined
in /usr/include/sysexits.h on Linux systems.

The codes and their meanings can also be found online at:

https://man.openbsd.org/sysexits.3
"""

from enum import Enum, IntEnum


class ExitCode(IntEnum):
    """Enumeration of exit codes.

    This class implements POSIX error codes, as found in
    /usr/include/sysexits.h on the Linux platform and
    described in the manual page for sysexit on BSD systems,
    including Darwin (macOS).

    Additional information is availble from The Linux Documnetation Project
    at http://tldp.org/LDP/abs/html/exitcodes.html
    """
    EX_SUCCESS = 0  # command exits successfully
    EX_GENERAL = 1  # catchall for general errors
    EX_USAGE = 64  # command line usage error
    EX_DATAERR = 65  # data format error
    EX_NOINPUT = 66  # cannot open input
    EX_NOUSER = 67  # addressee unknown
    EX_NOHOST = 68  # host name unknown
    EX_UNAVAILABLE = 69  # service unavailable
    EX_SOFTWARE = 70  # internal software error
    EX_OSERR = 71  # system error (e.g., can't fork)
    EX_OSFILE = 72  # critical OS file missing
    EX_CANTCREAT = 73  # can't create (user) output file
    EX_IOERR = 74  # input/output error
    EX_TEMPFAIL = 75  # temp failure; user is invited to retry
    EX_PROTOCOL = 76  # remote error in protocol
    EX_NOPERM = 77  # permission denied
    EX_CONFIG = 78  # configuration error
    EX_SIGINT = 130  # keyboard interrupt


class MXSError(Exception):
    """Base class for exceptions in this package."""
    pass



"""
MXS is the Math eXploration Solver.
Copyright (C) 2022  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

import os

from flask import Flask

from mxs.web import pattern


def create_app(test_config=None):
    # create and configure the app
    app = Flask(__name__, instance_relative_config=True)
    """
    __name__ is the name of the current Python module.
    The app needs to know where it’s located to set up some paths,
    and __name__ is a convenient way to tell it that.

    instance_relative_config=True tells the app that configuration files are
    relative to the instance folder. The instance folder is located outside
    the falsker package and can hold local data that shouldn’t be committed
    to version control, such as configuration secrets and the database file.
    """

    app.config.from_mapping(
        SECRET_KEY='dev',
    )

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)

    # ensure the instance folder exists
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    # a simple page that says hello
    @app.route('/hello')
    def hello():
        return 'Hello, World!'

    app.register_blueprint(pattern.bp)

    return app



"""
MXS is the Math eXploration Solver.
Copyright (C) 2022  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com


    mxs [args] [<command>] [<args>]

    mxs pattern --pv  # Reslove the position and value of an atom
                        within a line pattern
"""
import argparse

from mxs.__version__ import __version__
from .error import ExitCode
from .solvers import pattern_solver

# create the top-level parser
parser = argparse.ArgumentParser(
    prog='mxs',
    description='Use mxs to explore solutions for math problems.',
    conflict_handler='resolve',
    formatter_class=argparse.RawTextHelpFormatter,
    epilog='\n \n',
)

parser.add_argument(
    '-j', '--json',
    action='store_true',
    dest='json',
    help='Output the result set in JSON'
)

parser.add_argument(
    '--version',
    action='version',
    version='%(prog)s ' + f'version { __version__}'
)

# Add subparser below for each mxs command
subparsers = parser.add_subparsers(
    title='mxs commands',
)

# create the parser for the "pattern" command
parser_pattern = subparsers.add_parser(
    'pattern',
    formatter_class=argparse.RawTextHelpFormatter,
    help='solve a problem involving a pattern'
)

parser_pattern.add_argument(
    'problem',
    choices=[
        'pv',
    ],
    metavar='problem',
    help=('The type of pattern problem to solve.'
          ' Choices include pv (position/value).')
)

parser_pattern.set_defaults(func=pattern_solver)

args_optional = set(
    {
        'json': None
    }
)


def main():
    args = parser.parse_args() or None
    # logger.info(f"cli args: {vars(args)}")
    args_given = set(vars(args))
    check = args_given ^ args_optional

    if len(check) > 0:
        try:
            args.func(args)
        except SystemExit as e:
            return e.code
        else:
            return ExitCode.EX_SUCCESS
    else:
        parser.print_help()
        return ExitCode.EX_USAGE



"""
MXS is the Math eXploration Solver.
Copyright (C) 2022  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from setuptools import find_packages, setup

from mxs.__version__ import __version__

setup(
    name='mxs',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license='AGPLv3',
    zip_safe=False,
    install_requires=[
        'flask',
    ],
    entry_points={
        "console_scripts": [
            "mxs=mxs.cli:main",
        ],
    },
)



"""
MXS is the Math eXploration Solver.
Copyright (C) 2022  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from flask import (
    Blueprint, redirect, render_template, request, url_for, session
)

from mxs import api

bp = Blueprint('pattern', __name__, url_prefix='/pattern')


@bp.route('/answer')
def answer():
    return render_template('pattern/answer.html')


@bp.route('/pv', methods=('GET', 'POST'))
def reveal_position_value():
    if request.method == 'POST':
        pattern = request.form['pattern']
        char_per_line = request.form['char_per_line']
        target_line = request.form['target_line']
        target_char = request.form['target_char']

        target_position, value = api.reveal_position_value(
            pattern,
            char_per_line,
            target_line,
            target_char
        )

        session['answer'] = [
            {'label': 'Target Position', 'value': target_position},
            {'label': 'Value', 'value': value}
        ]

        return redirect(url_for('pattern.answer'))

    return render_template('pattern/pv.html')
